﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using dbfs;
using Xunit;

public class FolderTests
{
	[Fact]
	public void FoldersCacheIssue()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder1/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(2, result.Changes.Count);
			}

			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder1/", list);
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			// validates
			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(2, list.Count());
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);
			}
		}
	}

	[Fact]
	public void CreateSingleFolder()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
			commit.AddFolder("/folder/");

			var result = commit.Finish();

			Assert.NotNull(result);
			Assert.Equal(2, result.Changes.Count);

			// validate

			var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/");

			Assert.Equal(1, list.Count());
			Assert.Contains<string>("/folder/", list.Select(e => e.Path));
		}
	}

	[Fact]
	public void CreateMultipleFolders()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder/");

				Assert.NotNull(commit.Finish());
			}

			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder1/");
				commit.AddFolder("/folder/folder_inner/");
				commit.AddFolder("/folder2/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(3, result.Changes.Count);
			}

			// validate
			{
				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/")
					.Select(e => e.Path);

				Assert.Equal(3, list.Count());
				Assert.Contains<string>("/folder/", list);
				Assert.Contains<string>("/folder1/", list);
				Assert.Contains<string>("/folder2/", list);

				list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/folder/")
					.Select(e => e.Path);

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder/folder_inner/", list);
			}
		}
	}

	[Fact]
	public void RemoveFolder()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			// add
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder/");

				Assert.NotNull(commit.Finish());
			}

			// remove
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RemoveFolder("/folder/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			// validate

			var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/");

			Assert.Equal(0, list.Count());
		}
	}

	[Fact]
	public void RecreationOfFolder_Issue67()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS())
		{
			// add
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder/");

				Assert.NotNull(commit.Finish());
			}

			// remove
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RemoveFolder("/folder/");

				Assert.NotNull(commit.Finish());
			}

			// add again
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFolder("/folder/");

				var result = commit.Finish();

				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);

				// validate

				var list = dbfs.GetFoldersList(Utils.DEFAULT_BRANCH, "/");

				Assert.Equal(1, list.Count());
				Assert.Contains<string>("/folder/", list.Select(e => e.Path));
			}
		}
	}

	[Fact]
	public void RemoveAdd()
	{
		Utils.ClearDB();

		using (DBFS dbfs = Utils.CreateDBFS()) {
			// initialize
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.AddFile("/folder/file", p => new MemoryStream(Encoding.UTF8.GetBytes("file's first content")));

				var result = commit.Finish();
				Assert.NotNull(result);
				Assert.Equal(3, result.Changes.Count);
			}

			{
				Assert.True(dbfs.FolderExists(Utils.DEFAULT_BRANCH, "/folder/"));

				var files = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder/").Select(e => e.Path).ToList();
				Assert.Contains("/folder/file", files);
			}

			// check
			using (var commit = dbfs.StartCommit("", Utils.DEFAULT_USER, Utils.DEFAULT_BRANCH)) {
				commit.RemoveFolder("/folder/");
				commit.AddFile("/folder/file", p => new MemoryStream(Encoding.UTF8.GetBytes("file's second content")));

				var result = commit.Finish();
				Assert.NotNull(result);
				Assert.Equal(1, result.Changes.Count);
			}

			{
				Assert.True(dbfs.FolderExists(Utils.DEFAULT_BRANCH, "/folder/"));

				var files = dbfs.GetFilesList(Utils.DEFAULT_BRANCH, "/folder/").Select(e => e.Path).ToList();
				Assert.Contains("/folder/file", files);
			}
		}
	}
}
