﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.repository
{
	using System.Collections.Generic;
	using System.Linq;
	using dbfs.model;
	using dbfs.exceptions;

	class BranchManagement : RepositoryManagement
	{
		#region getters

		public Branch TryGet(int branch_id)
		{
			return _context.Branches.FirstOrDefault(e => e.id == branch_id);
		}

		/// <exception cref="BranchDoesNotExistException"></exception>
		public Branch Get(int branch_id)
		{
			var branch = TryGet(branch_id);

			if (branch == null)
				throw new BranchDoesNotExistException(branch_id);

			return branch;
		}

		public Branch TryGet(string name)
		{
			return _context.Branches.FirstOrDefault(e => e.name == name);
		}

		/// <exception cref="BranchDoesNotExistException"></exception>
		public Branch Get(string name)
		{
			var branch = TryGet(name);

			if (branch == null)
				throw new BranchDoesNotExistException(name);

			return branch;
		}

		public Branch TryGetByCommitId(int commit_id)
		{
			var bc = _context.BranchCommits
				.Where(e => e.commit_id == commit_id)
				.OrderBy(e => e.branch_id)
				.FirstOrDefault()
			;

			return bc == null ? null : bc.Branch;
		}

		/// <exception cref="BranchDoesNotExistException"></exception>
		public Branch GetByCommitId(int commit_id)
		{
			var branch = TryGetByCommitId(commit_id);

			if (branch == null)
				throw new BranchDoesNotExistException(null, commit_id);

			return branch;
		}

		public List<Branch> List()
		{
			return _context.Branches.ToList();
		}

		#endregion

		#region main manipulations

		/// <exception cref="BranchCreationFailedException"></exception>
		public Branch Create(string name, int user_id, string parent_branch)
		{
			if (string.IsNullOrEmpty(name))
				throw new BranchCreationFailedException("Branch name cannot be empty");

			var parent = Get(parent_branch);

			var new_branch_id = new System.Data.Entity.Core.Objects.ObjectParameter("new_branch_id", typeof(int));
			int result = _context.CreateBranch(name, parent.latest_commit_id, user_id, new_branch_id);

			if (result <= 0)
				throw new BranchCreationFailedException();

			Repository.Reload();

			return Get((int) new_branch_id.Value);
		}

		/// <exception cref="BranchRemoveFailedException"></exception>
		public void Remove(int branch_id)
		{
			int result = _context.RemoveBranch(branch_id);

			if (result <= 0)
				throw new BranchRemoveFailedException();

			Repository.Reload();
		}

		#endregion

		#region Lock/Unlock

		/// <param name="lock_timeout">In seconds</param>
		/// <exception cref="BranchLockedException"></exception>
		public void Lock(Branch branch, int lock_timeout, int user_id)
		{
			int result = _context.LockBranch(branch.id, user_id, lock_timeout);

			if (result > 0)
				_context.Entry(branch).Reload();
			else
				throw new BranchLockedException(branch.lock_release_date.Value, branch.lock_user_id.Value);
		}

		public void Unlock(Branch branch)
		{
			branch.lock_user_id = null;
			branch.lock_release_date = null;

			Repository.SaveChanges();
			_context.Entry(branch).Reload();
		}

		#endregion

		public bool Exists(string name)
		{
			return _context.Branches.Any(e => e.name == name);
		}
	}
}
