﻿/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.repository
{
	using System.Collections.Generic;
	using System.Linq;
	using dbfs.exceptions;
	using dbfs.model;
	using dbfs.utils;

	class RevisionManagement : RepositoryManagement
	{
		#region optimized getters

		/// <summary>If possible uses LatestRevisions to get requested revision.</summary>
		public FileRevision TryGetOptimized(Branch branch, string path, int commit_id = 0)
		{
			if (commit_id == 0 || branch.latest_commit_id == commit_id)
				return Repository.Latest.TryGet(branch.id, path)._(e => e.FileRevision);
			else
				return TryGet(branch.id, path, commit_id);
		}

		public FileRevision GetOptimized(Branch branch, string path, int commit_id = 0)
		{
			var revision = TryGetOptimized(branch, path, commit_id);

			if (revision == null)
				throw new RevisionNotFoundException(path, commit_id);

			return revision;
		}

		/// <summary>If possible uses LatestRevisions to get requested revision.</summary>
		/// <param name="only_folders">true - only folders, false - only files, null - folders and files</param>
		public List<FileRevision> ListOptimized(Branch branch, string directory, int commit_id = 0,
			bool? only_folders = null, bool skip_removed = false)
		{
			if (commit_id == 0 || branch.latest_commit_id == commit_id)
				return Repository.Latest.List(branch.id, directory, only_folders, skip_removed)
					.Select(e => e.FileRevision)
					.ToList();
			else
				return List(branch.id, directory, commit_id, only_folders, skip_removed);
		}

		/// <summary>If possible uses LatestRevisions to get requested revision.</summary>
		public List<FileRevision> AllNestedOptimized(Branch branch, string directory, int commit_id = 0,
			bool skip_removed = false)
		{
			if (commit_id == 0 || branch.latest_commit_id == commit_id)
				return Repository.Latest.AllNested(branch.id, directory, skip_removed)
					.Select(e => e.FileRevision)
					.ToList();
			else
				return AllNested(branch.id, directory, commit_id, skip_removed);
		}

		#endregion

		#region getters

		public FileRevision TryGet(int branch_id, string path, int commit_id = 0)
		{
			var revisions = _context.FileRevisions.Where(e => e.File.path == path);

			return _context.BranchCommits
				.Where(e => e.branch_id == branch_id)
				.If(commit_id != 0, q => q.Where(e => e.commit_id <= commit_id))
				.Join(revisions, e => e.commit_id, e => e.commit_id, (a, r) => r)
				.OrderByDescending(e => e.commit_id)
				.FirstOrDefault()
			;
		}

		/// <param name="only_folders">true - only folders, false - only files, null - folders and files</param>
		public List<FileRevision> List(int branch_id, string directory, int commit_id = 0,
			bool? only_folders = null, bool skip_removed = false)
		{
			return _context.BranchCommits
				.Where(e => e.branch_id == branch_id)
				.If(commit_id != 0, q => q.Where(e => e.commit_id <= commit_id))
				.Join(_context.FileRevisions, e => e.commit_id, e => e.commit_id, (a, r) => r)
				.Where(e => e.File.folder == directory)
				.If(only_folders.HasValue, q => q.Where(e => e.File.is_directory == only_folders.Value))
				.GroupBy(e => e.file_id)
				.Select(e => e.OrderByDescending(r => r.commit_id).FirstOrDefault())
				.If(skip_removed, q => q.SkipRemoved())
				.ToList()
			;
		}

		public List<FileRevision> AllNested(int branch_id, string directory, int commit_id = 0,
			bool skip_removed = false)
		{
			var revisions = _context.FileRevisions.Where(e => e.File.path.StartsWith(directory));

			return _context.BranchCommits
				.Where(e => e.branch_id == branch_id)
				.If(commit_id != 0, q => q.Where(e => e.commit_id <= commit_id))
				.Join(revisions, e => e.commit_id, e => e.commit_id, (a, r) => r)
				.GroupBy(e => e.file_id)
				.Select(e => e.OrderByDescending(r => r.commit_id).FirstOrDefault())
				.If(skip_removed, q => q.SkipRemoved())
				.ToList()
			;
		}

		#endregion

		public bool Exists(int branch_id, string path, int commit_id = 0)
		{
			var revision = _context.BranchCommits
				.Where(e => e.branch_id == branch_id)
				.If(commit_id != 0, q => q.Where(e => e.commit_id <= commit_id))
				.Join(_context.FileRevisions, e => e.commit_id, e => e.commit_id, (a, r) => r)
				.Where(e => e.File.path == path)
				.OrderByDescending(e => e.commit_id)
				.FirstOrDefault()
			;

			return revision != null && !revision.IsRemoved();
		}
	}
}
