/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

namespace dbfs.dbfscmd
{
	using System;
	using System.Collections.Generic;
	using ManyConsole;

	public class ListCommand : ConsoleCommand
	{
		public ListCommand()
		{
			this.SkipsCommandSummaryBeforeRunning();
			this.IsCommand("list", "get files/folders lists");

			this.HasOption("files", "", v => _is_files = v != null);
			this.HasOption("folders", "", v => _is_folders = v != null);
			this.HasOption("e|extension=", "provide file extension to filter by", v => _ext = v);
			this.HasOption<int>("c|commit=", "provide commit id", v => this._commit_id = v);
			this.HasOption<string>("b|branch=", "provide branch name", v => this._branch = v);

			this.HasAdditionalArguments(1, "<virtual-path>");
		}

		private bool _is_files = false;
		private bool _is_folders = false;
		private string _ext;
		private int? _commit_id;
		private string _branch;

		public override int Run(string[] remainingArguments)
		{
			this._checkArguments(remainingArguments);

			string path = remainingArguments[0];
			var dbfs = Program.GetDBFS();

			if (this._commit_id.HasValue) {
				if (this._is_folders) {
					this._listFolders(dbfs.GetFoldersList(this._commit_id.Value, path));
				}
				if (this._is_files) {
					this._listFiles(dbfs.GetFilesList(this._commit_id.Value, path /*, this._ext*/));
				}
			}
			else {
				if (this._is_folders)
					this._listFolders(dbfs.GetFoldersList(this._branch, path));
				if (this._is_files)
					this._listFiles(dbfs.GetFilesList(this._branch, path));
			}

			return 0;
		}

		private void _checkArguments(string[] remainingArguments)
		{
			if (!this._is_folders && !this._is_files)
				throw new ConsoleHelpAsException("at least one argument required: --folders, --files");

			if (!this._commit_id.HasValue && String.IsNullOrEmpty(this._branch))
				throw new ConsoleHelpAsException("one argument required: --commit, --branch");

			if (this._commit_id.HasValue && !String.IsNullOrEmpty(this._branch))
				throw new ConsoleHelpAsException("--commit and --branch cannot appear both at the same time");

			if (!this._is_files && !String.IsNullOrEmpty(this._ext))
				throw new ConsoleHelpAsException("extension filtering works only for files");
		}

		private void _listFolders(IEnumerable<dbfs.Folder> folders)
		{
			foreach (var folder in folders) {
				Console.WriteLine(String.Format("{0}\t\t\t{1}\t\t\t{2}", folder.Path, folder.Date, folder.UserName));
			}
		}

		private void _listFiles(IEnumerable<dbfs.File> files)
		{
			foreach (var file in files) {
				Console.WriteLine(file.Path);
			}
		}
	}
}
