/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections;
using System.ComponentModel;
using System.Globalization;

namespace dbfs.utils
{
	public static class CastUtils
	{
		/// <summary>
		/// Check if the value is null. Nullable without value and DBNull are treated as null.
		/// </summary>
		public static bool IsNull(this object v)
		{
			return object.ReferenceEquals(v, null) || v is DBNull;
		}

		public static bool IsNumberType(this Type type)
		{
			TypeCode code = Type.GetTypeCode(type);

			return code == TypeCode.Byte
				|| code == TypeCode.SByte
				|| code == TypeCode.Int16
				|| code == TypeCode.UInt16
				|| code == TypeCode.Int32
				|| code == TypeCode.UInt32
				|| code == TypeCode.Int64
				|| code == TypeCode.UInt64
				|| code == TypeCode.Single
				|| code == TypeCode.Double
				|| code == TypeCode.Decimal;
		}

		/// <summary>
		/// Check if the value is number (works for Nullable types as well).
		/// </summary>
		/// <param name="v"></param>
		/// <returns></returns>
		public static bool IsNumber(this object v)
		{
			return !v.IsNull() && IsNumberType(v.GetType());
		}

		public static bool IsFloat(this object v)
		{
			return v is double || v is float || v is decimal;
		}

		public static bool IsZero(dynamic v)
		{
			return IsNumber(v) && v == 0;
		}

		public static bool IsNullableType(this Type type)
		{
			return type.IsGenericType
				&& type.GetGenericTypeDefinition().Equals(typeof(Nullable<>));
		}

		public static bool IsArray(this object v)
		{
			return v is IEnumerable && !(v is string) && !(v is IDictionary);
		}

		public static string GetTypeName(this object obj)
		{
			return !IsNull(obj) ? obj.GetType().Name : "null";
		}

		public static int ToInt(this string text)
		{
			if (string.IsNullOrEmpty(text))
				return 0;

			return int.Parse(
				text.Replace("_", ""),
				CultureInfo.InvariantCulture
			);
		}

		public static double ToDouble(this string text)
		{
			if (string.IsNullOrEmpty(text))
				return 0.0;

			return double.Parse(text, CultureInfo.InvariantCulture);
		}

		public static uint ToUInt(this object obj)
		{
			try
			{
				ulong lval = Convert.ToUInt64(obj);
				unchecked { return (uint)lval; }
			}
			catch
			{
				return 0;
			}
		}

		public static bool ToBool(this object obj)
		{
			if (obj.IsNull())
				return false;

			if (obj is bool)
				return (bool)obj;

			if (obj is string)
				return ((string)obj).Length != 0;

			if (obj is IEnumerable)
				return !((IEnumerable)obj).IsEmpty();

			if (IsZero(obj))
				return false;

			return true;
		}

		public static bool MayCast(this object value, Type type)
		{
			if (type == null)
				throw new ArgumentNullException("type");

			if (value.IsNull())
				return !type.IsValueType || type.IsNullableType();

			Type valueType = value.GetType();

			return valueType.Equals(type) || valueType.IsSubclassOf(type) || type.IsAssignableFrom(valueType);
		}

		public static bool MayCast<T>(this object value)
		{
			return MayCast(value, typeof(T));
		}

		public static object ChangeType(this object value, Type type)
		{
			if (type == null)
				throw new ArgumentNullException("type");

			if (type.IsNullableType())
			{
				if (value.IsNull())
					return null;

				type = new NullableConverter(type).UnderlyingType;
			}

			return Convert.ChangeType(value, type, CultureInfo.InvariantCulture);
		}

		public static T ChangeType<T>(this object value)
		{
			return (T)ChangeType(value, typeof(T));
		}

		public static object CastOrConvert(this object value, Type type)
		{
			if (MayCast(value, type))
				return value.IsNull() ? null : value;

			return ChangeType(value, type);
		}

		public static T CastOrConvert<T>(this object value)
		{
			return (T)CastOrConvert(value, typeof(T));
		}

		public static bool TryCastOrConvert(this object value, Type type, out object result)
		{
			try
			{
				result = CastOrConvert(value, type);
				return true;
			}
			catch
			{
				result = null;
				return false;
			}
		}

		public static bool TryCastOrConvert<T>(this object value, out T result)
		{
			try
			{
				result = CastOrConvert<T>(value);
				return true;
			}
			catch
			{
				result = default(T);
				return false;
			}
		}
	}
}
