/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace dbfs.utils
{
	public static class EnumerableUtils
	{
		/// <summary>
		/// Check if the sequence is empty.
		/// </summary>
		public static bool IsEmpty(this IEnumerable sequence)
		{
			sequence.ThrowIfNull("sequence");

			foreach (var el in sequence)
				return false;

			return true;
		}

		/// <summary>
		/// Try getting an element from the sequence at the specified index.
		/// </summary>
		/// <param name="index">Index to get the item at</param>
		/// <param name="item">The result (if the sequence is large enough)</param>
		/// <param name="count">The index of the item if found, or the sequence length if not found</param>
		/// <returns>true if found</returns>
		public static bool TryGetItem<T>(this IEnumerable<T> sequence, int index, out T item, out int count)
		{
			sequence.ThrowIfNull("sequence");

			if (index < 0)
				throw new ArgumentOutOfRangeException("index");

			item = default(T);
			IList<T> list = sequence as IList<T>;

			if (list != null)
			{
				if (index >= list.Count)
				{
					count = list.Count;
					return false;
				}

				item = list[index];
				count = index;

				return true;
			}

			count = 0;

			foreach (var el in sequence)
			{
				if (index == count)
				{
					item = el;
					return true;
				}

				++count;
			}

			return false;
		}

		/// <summary>
		/// Get index of the element in the sequence.
		/// </summary>
		/// <param name="item">The element to search for (default comparer is used)</param>
		/// <param name="index">The index of matching element in the sequence, or the sequence length if there is no matching element</param>
		/// <returns>true if found</returns>
		public static bool FindIndex<T>(this IEnumerable<T> sequence, T item, out int index)
		{
			sequence.ThrowIfNull("sequence");

			IList<T> list = sequence as IList<T>;

			if (list != null)
			{
				index = list.IndexOf(item);

				if (index >= 0)
					return true;

				index = list.Count;
				return false;
			}

			var comparer = EqualityComparer<T>.Default;

			index = 0;

			foreach (var test in sequence)
			{
				if (comparer.Equals(test, item))
					return true;

				++index;
			}

			return false;
		}

		public static void CopyToArray<T>(this ICollection<T> collection, IList array, int arrayIndex)
		{
			if (array == null)
				throw new ArgumentNullException("array");

			if (arrayIndex < 0)
				throw new ArgumentOutOfRangeException("arrayIndex");

			foreach (T item in collection)
			{
				if (array.Count <= arrayIndex)
					throw new ArgumentException("There is not enough room to copy elements", "array");

				array[arrayIndex++] = item;
			}
		}

		public static IEnumerable<TResult> FullOuterJoin<TOuter, TInner, TKey, TResult>(
			this IEnumerable<TOuter> outer,
			IEnumerable<TInner> inner,
			Func<TOuter, TKey> outerKeySelector,
			Func<TInner, TKey> innerKeySelector,
			Func<TOuter, TInner, TResult> resultSelector,
			IEqualityComparer<TKey> comparer)
			where TInner : class
			where TOuter : class
		{
			outer.ThrowIfNull("outer");
			inner.ThrowIfNull("inner");
			outerKeySelector.ThrowIfNull("outerKeySelector");
			resultSelector.ThrowIfNull("resultSelector");
			comparer.ThrowIfNull("comparer");

			ILookup<TKey, TInner> innerLookup = inner.ToLookup(innerKeySelector, comparer);
			ILookup<TKey, TOuter> outerLookup = outer.ToLookup(outerKeySelector, comparer);

			IEnumerable<TResult> innerJoinItems = inner
				.Where(innerItem => !outerLookup.Contains(innerKeySelector(innerItem)))
				.Select(innerItem => resultSelector(null, innerItem));

			return outer
				.SelectMany(
					outerItem => {
						IEnumerable<TInner> innerItems = innerLookup[outerKeySelector(outerItem)];
						return innerItems.Any() ? innerItems : new TInner[] { null };
					},
					resultSelector)
				.Concat(innerJoinItems)
			;
		}

		public static IEnumerable<TResult> FullOuterJoin<TOuter, TInner, TKey, TResult>(
			this IEnumerable<TOuter> outer,
			IEnumerable<TInner> inner,
			Func<TOuter, TKey> outerKeySelector,
			Func<TInner, TKey> innerKeySelector,
			Func<TOuter, TInner, TResult> resultSelector)
			where TInner : class
			where TOuter : class
		{
			return FullOuterJoin(
				outer, inner, outerKeySelector, innerKeySelector, resultSelector,
				EqualityComparer<TKey>.Default
			);
		}

		public static IEnumerable<TResult> LeftJoin<TOuter, TInner, TKey, TResult>(
			this IEnumerable<TOuter> outer,
			IEnumerable<TInner> inner,
			Func<TOuter, TKey> outerKeySelector,
			Func<TInner, TKey> innerKeySelector,
			Func<TOuter, TInner, TResult> resultSelector,
			IEqualityComparer<TKey> keyComparer)
			where TInner : class
			where TOuter : class
		{
			outer.ThrowIfNull("outer");
			inner.ThrowIfNull("inner");
			outerKeySelector.ThrowIfNull("outerKeySelector");
			resultSelector.ThrowIfNull("resultSelector");
			keyComparer.ThrowIfNull("comparer");

			return outer
				.GroupJoin(
					inner, outerKeySelector, innerKeySelector,
					(outerItem, innerItems) => new { Outer = outerItem, InnerItems = innerItems },
					keyComparer)
				.SelectMany(
					group => group.InnerItems.DefaultIfEmpty(),
					(group, innerItem) => resultSelector(group.Outer, innerItem))
			;
		}

		public static IEnumerable<TResult> LeftJoin<TOuter, TInner, TKey, TResult>(
			this IEnumerable<TOuter> outer,
			IEnumerable<TInner> inner,
			Func<TOuter, TKey> outerKeySelector,
			Func<TInner, TKey> innerKeySelector,
			Func<TOuter, TInner, TResult> resultSelector)
			where TOuter : class
			where TInner : class
		{
			return LeftJoin(
				outer, inner, outerKeySelector, innerKeySelector, resultSelector,
				EqualityComparer<TKey>.Default
			);
		}

		public static bool SequenceEqual<T>(this IEnumerable<T> first, IEnumerable<T> second, Func<T,T,bool> comparator)
		{
			using (var _first = first.GetEnumerator())
			using (var _second = second.GetEnumerator()) {
				for (;;) {
					bool a = _first.MoveNext();
					bool b = _second.MoveNext();

					if (a && b) {
						if (!comparator(_first.Current, _second.Current))
							return false;
					}
					else return a == b;
				}
			}
		}

		/// <summary>
		/// Execute action for every element.
		/// </summary>
		/// <returns>sequence itself</returns>
		public static IEnumerable<T> ForEach<T>(this IEnumerable<T> sequence, Action<T> action)
		{
			sequence.ThrowIfNull("sequence");

			foreach (var e in sequence) {
				action(e);
			}

			return sequence;
		}

		/// <summary>
		/// Execute action for every element. Execution can be stopped by invoking stop action (second argument).
		/// </summary>
		/// <typeparam name="T"></typeparam>
		/// <param name="sequence"></param>
		/// <param name="action"></param>
		/// <returns></returns>
		public static IEnumerable<T> ForEach<T>(this IEnumerable<T> sequence, Action<T, Action> action)
		{
			sequence.ThrowIfNull("sequence");

			bool stop = false;
			Action stop_ = () => { stop = true; };

			foreach (var e in sequence) {
				if (stop) break;

				action(e, stop_);
			}

			return sequence;
		}
	}
}
