/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace dbfs.utils
{
	public static class FileUtils
	{
		private const char SLASH_CHR = '/';
		private const char BSLASH_CHR = '\\';
		private const char DOT_CHR = '.';
		private const string SLASH_STR = "/";
		private const string DBL_SLASH_STR = "//";
		private const string DBL_BSLASH_STR = @"\\";
		private const string PARENT_STR = "../";
		private const string PARENT_SIGNATURE = "/..";
		private const string URL_PREFIX = "://";

		public static string BaseDirectory
		{
			get { return NormalizePath(AppDomain.CurrentDomain.BaseDirectory); }
		}

		public static bool IsUrl(string path)
		{
			return path != null && path.Contains(URL_PREFIX);
		}

		public static bool IsAbsolutePath(string path)
		{
			return path != null && (IsUrl(path) || Path.IsPathRooted(path));
		}

		public static string NormalizePath(string path)
		{
			string norm_path = (path ?? "").Replace(BSLASH_CHR, SLASH_CHR).TrimEnd(SLASH_CHR);
			int prev_pos = -1;

			while (true)
			{
				int pos = norm_path.IndexOf(SLASH_CHR, prev_pos + 1);

				if (pos < 0 || pos == norm_path.Length - 1)
					break;

				int end = norm_path.IndexOf(SLASH_CHR, pos + 1);

				if (end < 0)
					end = norm_path.Length;

				if (prev_pos >= 0 && norm_path.Substring(pos, end - pos) == PARENT_SIGNATURE)
				{
					norm_path = norm_path.Remove(prev_pos + 1, end - prev_pos);

					for (--prev_pos; prev_pos >= 0 && norm_path[prev_pos] != SLASH_CHR; --prev_pos)
						;
				}
				else
				{
					prev_pos = pos;
				}
			}

			if (norm_path.StartsWith(DBL_SLASH_STR)) //keep network prefix
				norm_path = DBL_BSLASH_STR + norm_path.Substring(2);

			return norm_path;
		}

		public static string GetFullPath(string path)
		{
			if (IsAbsolutePath(path))
				return NormalizePath(path);

			return NormalizePath(Path.GetFullPath(Path.Combine(BaseDirectory, path ?? "")));
		}

		public static string CombinePath(string base_dir, string tail_path)
		{
			if (IsAbsolutePath(tail_path))
				return NormalizePath(tail_path);

			return NormalizePath(Path.Combine(base_dir ?? "", tail_path ?? ""));
		}

		public static string GetDirectoryName(string file_path)
		{
			string path = NormalizePath(file_path);
			int pos = path.LastIndexOf(SLASH_CHR);
			return pos >= 0 ? path.Substring(0, pos) : "";
		}

		public static string GetLastDirectoryName(string file_path)
		{
			string path = GetDirectoryName(file_path);
			return path.Substring(path.LastIndexOf(SLASH_CHR) + 1);
		}

		public static string GetFileName(string file_path)
		{
			string path = NormalizePath(file_path);
			return path.Substring(path.LastIndexOf(SLASH_CHR) + 1);
		}

		public static string ChangeExtension(string file_path, string new_ext_with_dot)
		{
			string path = NormalizePath(file_path);
			int slash_pos = path.LastIndexOf(SLASH_CHR);
			int dot_pos = path.LastIndexOf(DOT_CHR);

			if (dot_pos < slash_pos || dot_pos < 0)
				return string.Concat(path, new_ext_with_dot);

			return string.Concat(path.Substring(0, dot_pos), new_ext_with_dot);
		}

		/// <summary>Returns first relative path or path itself.</summary>
		public static string GetRelativePathFirst(string path, params string[] to_dirs)
		{
			foreach (var dir in to_dirs) {
				string relative_path = GetRelativePath(path, dir);

				if (!IsAbsolutePath(relative_path)) return relative_path;
			}

			return path;
		}

		public static bool PathStartsWithDirectory(string target_path, string dir)
		{
			return string.IsNullOrEmpty(dir)
				|| NormalizePath(target_path)
					.StartsWith(NormalizePath(dir).TrimEnd(SLASH_CHR) + SLASH_CHR);
		}

		public static string GetRelativePath(string path, string to_dir)
		{
			string tgt_path = GetFullPath(path);
			string my_dir = GetFullPath(to_dir);

			if (IsUrl(tgt_path))
				return tgt_path;

			string[] tgt_parts = tgt_path.Split(SLASH_CHR);
			string[] my_parts = my_dir.Split(SLASH_CHR);

			if (tgt_parts.Length == 0 || my_parts.Length == 0 || tgt_parts[0] != my_parts[0])
				return tgt_path;

			StringBuilder rel_path = new StringBuilder();
			int i = 0;

			for (; i < my_parts.Length; ++i)
			{
				if (i >= tgt_parts.Length || tgt_parts[i] != my_parts[i])
				{
					for (int j = i; j < my_parts.Length; ++j)
						rel_path.Append(PARENT_STR);

					break;
				}
			}

			for (int j = i; j < tgt_parts.Length; ++j)
			{
				if (j != i)
					rel_path.Append(SLASH_CHR);

				rel_path.Append(tgt_parts[j]);
			}

			return rel_path.ToString();
		}

		/// <summary>Returns first relative path or path itself.</summary>
		public static string GetFirstRelativePath(string path, params string[] to_dirs)
		{
			foreach (var dir in to_dirs)
			{
				string relative_path = GetRelativePath(path, dir);

				if (!IsAbsolutePath(relative_path)) return relative_path;
			}

			return path;
		}

		public static string GetRelativePathThatStartsWith(string path, params string[] dirs)
		{
			return GetRelativePathAndNumberThatStartsWith(path, dirs).Item1;
		}

		/// <summary>
		/// Returns Tuple with relative path and number of that path.
		/// Number starts from 1. 0 means path isn't relative to any of the directories.
		/// </summary>
		public static Tuple<string, int> GetRelativePathAndNumberThatStartsWith(string path, params string[] dirs)
		{
			int i = 1;

			foreach (var dir in dirs) {
				if (FileUtils.PathStartsWithDirectory(path, dir))
					return new Tuple<string,int>(FileUtils.GetRelativePath(path, dir), i);

				++i;
			}

			return new Tuple<string, int>(path, 0);
		}

		public static string GetCommonPath(string path1, string path2)
		{
			if (string.IsNullOrEmpty(path1) || string.IsNullOrEmpty(path2))
				return "";

			path1 = GetFullPath(path1);
			path2 = GetFullPath(path2);

			if (path1 == path2)
				return path1;

			string[] parts1 = path1.Split(SLASH_CHR);
			string[] parts2 = path2.Split(SLASH_CHR);

			int len = Math.Min(parts1.Length, parts2.Length);

			for (int i = 0; i <= len; ++i)
			{
				if (i == len || parts1[i] != parts2[i])
					return string.Join(SLASH_STR, parts1, 0, i);
			}

			return "";
		}

		public static string GetCommonPath(IEnumerable<string> paths)
		{
			string common = null;

			foreach (string path in paths)
			{
				if (common == null)
					common = GetFullPath(path);
				else
					common = GetCommonPath(common, path);
			}

			return common ?? "";
		}

		public static string StripFileName(string filename)
		{
			string stripped = (filename ?? "").Replace("..", "_");

			foreach (char ch in Path.GetInvalidFileNameChars())
				stripped = stripped.Replace(ch, '_');

			return stripped;
		}

        public static bool IsValidFileName(string filename)
        {
            if (filename.IndexOf('#') > -1)
                return false;

            foreach (char ch in Path.GetInvalidFileNameChars())
                if (filename.IndexOf(ch) > -1)
                    return false;

            return true;
        }

		public static string CreateDir(string path)
		{
			string dir = GetFullPath(path);

			if (!Directory.Exists(dir))
				Directory.CreateDirectory(dir);

			return dir;
		}

		public static string CreateDir(string base_dir, string new_dir)
		{
			return CreateDir(CombinePath(base_dir, new_dir));
		}

		public static bool DirectoryExists(string path)
		{
			return Directory.Exists(GetFullPath(path));
		}

		public static bool FileExists(string path)
		{
			return File.Exists(GetFullPath(path));
		}

		public static void DeleteEmptyDirectory(string path)
		{
			Directory.Delete(path);
		}

		public static void SaveToStream(this Stream in_stream, Stream out_stream)
		{
			int count;
			byte[] buf = new byte[4 * 1024 * 1024]; // 4MB buffer
			long original_pos = in_stream.CanSeek ? in_stream.Position : 0;

			try
			{
				while ((count = in_stream.Read(buf, 0, buf.Length)) != 0)
				{
					out_stream.Write(buf, 0, count);
				}
			}
			finally
			{
				if (in_stream.CanSeek)
					in_stream.Position = original_pos;
			}
		}

		public static void SaveToFile(this Stream in_stream, string filename)
		{
			using (var file = new FileStream(GetFullPath(filename), FileMode.Create))
			{
				in_stream.SaveToStream(file);
			}
		}

		public static string GetAsString(this Stream stream)
		{
			using (StreamReader reader = new StreamReader(stream))
			{
				return reader.ReadToEnd();
			}
		}
	}
}
