/*
	DbFS.NET
    Copyright (C) 2016 SCAND Ltd, svg@scand.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

using System;
using System.IO;

using log4net;
using log4net.Core;
using log4net.Config;

namespace dbfs.utils
{
	public class Logger
	{
		private static readonly Level MessageLevel = new Level(35000, "MESSAGE");

		static Logger()
		{
			LogManager.GetRepository().LevelMap.Add(MessageLevel);

			var config = new FileInfo(FileUtils.GetFullPath("log.config"));

			if (config.Exists)
				XmlConfigurator.ConfigureAndWatch(config);
			else
				XmlConfigurator.Configure();
		}

		public Logger(string name)
		{
			_log = LogManager.GetLogger(name);
		}

		public void Message(string message, params object[] args)
		{
			_log.Logger.Log(
				System.Reflection.MethodBase.GetCurrentMethod().DeclaringType,
				MessageLevel, String.Format(message, args), null
			);
		}

		public void Info(string message, params object[] args)
		{
			_log.InfoFormat(message, args);
		}

		public void Warn(string message, params object[] args)
		{
			_log.WarnFormat(message, args);
		}

		public void Error(string message, params object[] args)
		{
			_log.ErrorFormat(message, args);
		}

		public void Error(string message, Exception e)
		{
			_log.Error(message, e);
		}

		public void Fatal(string message, params object[] args)
		{
			_log.FatalFormat(message, args);
		}

		public void Fatal(string message, Exception e)
		{
			_log.Fatal(message, e);
		}

		public void Debug(string message, params object[] args)
		{
			_log.DebugFormat(message, args);
		}

		public void Flush()
		{
			log4net.Repository.ILoggerRepository repo = _log.Logger.Repository;

			if (repo != null)
			{
				foreach (log4net.Appender.IAppender appender in repo.GetAppenders())
				{
					var buffered = appender as log4net.Appender.BufferingAppenderSkeleton;

					if (buffered != null)
						buffered.Flush();
				}
			}
		}

		private readonly ILog _log;
	}

	public static class Log
	{
		public static void Message(string message, params object[] args)
		{
			Default.Message(message, args);
		}

		public static void Info(string message, params object[] args)
		{
			Default.Info(message, args);
		}

		public static void Warn(string message, params object[] args)
		{
			Default.Warn(message, args);
		}

		public static void Error(string message, params object[] args)
		{
			Default.Error(message, args);
		}

		public static void Error(string message, Exception e)
		{
			Default.Error(message, e);
		}

		public static void Fatal(string message, params object[] args)
		{
			Default.Fatal(message, args);
		}

		public static void Fatal(string message, Exception e)
		{
			Default.Fatal(message, e);
		}

		public static void Debug(string message, params object[] args)
		{
			Default.Debug(message, args);
		}

		public static void Flush()
		{
			Default.Flush();
		}

		private static Logger Default
		{
			get
			{
				//TODO: make it thread safe
				if (_logger == null)
					_logger = new Logger("Default");

				return _logger;
			}
		}

		private static Logger _logger;
	}
}
